////////////////////////////////////////////////////////////////////////////////
// CONFIDENTIAL and PROPRIETARY software of Magewell Electronics Co., Ltd.
// Copyright (c) 2011-2019 Magewell Electronics Co., Ltd. (Nanjing)
// All rights reserved.
// This copyright notice MUST be reproduced on all authorized copies.
////////////////////////////////////////////////////////////////////////////////

#import "ViewController.h"
#import <mach/mach.h>
#import <mach/mach_time.h>
#import <mach-o/dyld.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <queue>

struct PixelBufferFrame {
    PixelBufferFrame(): pixelBuffer(NULL), timestamp(0), discard(NULL), param(NULL) {
    }
    
    ~PixelBufferFrame() {
        if (discard) { discard(this, param); }
    }
    
    CVPixelBufferRef pixelBuffer;
    uint64_t timestamp;
    
    void (*discard)(PixelBufferFrame *frame, void *param);
    void *param;
};

static void releasePixelBufferFrame(PixelBufferFrame *frame, void *param) {
    ViewController *viewCtrl = (__bridge ViewController*)param;
    
    if (frame && frame->pixelBuffer) {
        [[viewCtrl capturePixelFrameLock] lock];
        ((std::queue<CVPixelBufferRef> *)[viewCtrl capturePixelFrameQueue])->push(frame->pixelBuffer);
        [[viewCtrl capturePixelFrameLock] unlock];
    }
}

static uint64_t getTimestamp() {
    uint64_t time = mach_absolute_time();
    static mach_timebase_info_data_t sTimebaseInfo;
    
    if (sTimebaseInfo.denom == 0) {
        mach_timebase_info(&sTimebaseInfo);
    }
    
    long double ret = 1.L * time / 1000 * sTimebaseInfo.numer / sTimebaseInfo.denom;
    return ret / 1000;
}

static void* onVideoCaptureThreadProc(void *param) {
    ViewController *viewCtrl = (__bridge ViewController*)param;
    
    [viewCtrl onVideoCaptureProc];
    
    return NULL;
}

static void* onAudioCaptureThreadProc(void *param) {
    ViewController *viewCtrl = (__bridge ViewController*)param;
    
    [viewCtrl onAudioCaptureProc];
    
    return NULL;
}

static void* onVideoEncodeThreadProc(void *param) {
    ViewController *viewCtrl = (__bridge ViewController*)param;
    
    [viewCtrl onVideoEncodeProc];
    
    return NULL;
}

static void onVTBoxEncodeCallback(void * user_ptr, const uint8_t* nalus[], uint32_t nalu_lens[],uint32_t nalu_num, mw_venc_frame_info_t *p_frame_info) {
    ViewController *viewCtrl = (__bridge ViewController*)user_ptr;
    
    [viewCtrl onVTBoxEncodeFrame:nalus nalu_lens:nalu_lens nalu_num:nalu_num frame_info:p_frame_info];
}

static void onAacEncodeCallback(uint8_t *frame, size_t frame_size, uint64_t timestamp, uint32_t seq, void *param) {
    ViewController *viewCtrl = (__bridge ViewController*)param;
    [viewCtrl onAacEncodeFrame:frame size:frame_size timestamp:timestamp];
}

#define MAX_VIDEO_ENCODE_BUFFER_FRAMES 6
#define MAX_VIDEO_CAPTURE_BUFFER_FRAMES (MAX_VIDEO_ENCODE_BUFFER_FRAMES + 2)

@implementation ViewController
- (void)onVideoCaptureFrame:(BYTE *)pBuffer bufferLen:(long)bufferLen pixelBufferFrame:(std::shared_ptr<PixelBufferFrame>) frame  {
    //render
    if (self.viewEnable) {
        CMSampleTimingInfo timing = {kCMTimeInvalid, kCMTimeInvalid, kCMTimeInvalid};
        CMVideoFormatDescriptionRef videoInfo = NULL;
        CVReturn result = CMVideoFormatDescriptionCreateForImageBuffer(NULL, frame->pixelBuffer, &videoInfo);

        CMSampleBufferRef sampleBuffer = NULL;
        result = CMSampleBufferCreateForImageBuffer(kCFAllocatorDefault, frame->pixelBuffer, true, NULL, NULL, videoInfo, &timing, &sampleBuffer);
        CFRelease(videoInfo);

        CFArrayRef attachments = CMSampleBufferGetSampleAttachmentsArray(sampleBuffer, YES);
        CFMutableDictionaryRef dict = (CFMutableDictionaryRef)CFArrayGetValueAtIndex(attachments, 0);
        CFDictionarySetValue(dict, kCMSampleAttachmentKey_DisplayImmediately, kCFBooleanTrue);

        if (self.videoLayer) {
            [self.videoLayer enqueueSampleBuffer:(CMSampleBufferRef)sampleBuffer];
        }
        
        CFRelease(sampleBuffer);
    }
    
    //encode
    [self.vtEncLock lock];
    if (self.vtEnc) {
        ((std::queue<std::shared_ptr<PixelBufferFrame> > *)self.encPixelFrameQueue)->push(frame);
        while(((std::queue<std::shared_ptr<PixelBufferFrame> > *)self.encPixelFrameQueue)->size() > MAX_VIDEO_ENCODE_BUFFER_FRAMES) {
            ((std::queue<std::shared_ptr<PixelBufferFrame> > *)self.encPixelFrameQueue)->pop();
        }
    }
    [self.vtEncLock unlock];

    //capture fps
    {
        uint64_t now = getTimestamp();
        if (self.lastTimestamp == 0) {
            self.lastTimestamp = now;
            self.videoFrameCount = 0;
        } else {
            self.videoFrameCount++;
        }

        if (now - self.lastTimestamp >= 1000) {
            self.captureFps = self.videoFrameCount * 1000.0 / (now - self.lastTimestamp);
            self.lastTimestamp = now;
            self.videoFrameCount = 0;
        }
    }
}

- (void)onVideoCaptureProc {
    CVPixelBufferRef pixelBuffer[MAX_VIDEO_CAPTURE_BUFFER_FRAMES] = {NULL};
    NSDictionary *pixelAttributes = @{(NSString*)kCVPixelBufferIOSurfacePropertiesKey:@{}};
    for (int i = 0; i < sizeof(pixelBuffer)/sizeof(pixelBuffer[0]); i++) {
        CVReturn result = CVPixelBufferCreate(kCFAllocatorDefault,
                                              self.width,
                                              self.height,
                                              self.pixelFormat,
                                              (__bridge CFDictionaryRef)(pixelAttributes),
                                              &pixelBuffer[i]);
        
        if (kCVReturnSuccess == result) {
            ((std::queue<CVPixelBufferRef> *)self.capturePixelFrameQueue)->push(pixelBuffer[i]);
        }
    }

    // Wait Events
    MWCAP_PTR hCaptureEvent    = MWCreateEvent();
    MWCAP_PTR hTimerEvent      = MWCreateEvent();

    MW_RESULT xr;
    do {
        xr = MWStartVideoCapture(self.hChannel, hCaptureEvent);
        if (xr != MW_SUCCEEDED)
            break;

        HTIMER hTimerNotify = MWRegisterTimer(self.hChannel, hTimerEvent);
        if (hTimerNotify == 0)
            break;

        LONGLONG llBegin = 0LL;
        xr = MWGetDeviceTime(self.hChannel, &llBegin);
        if (xr != MW_SUCCEEDED)
            break;

        LONGLONG llExpireTime = llBegin;
        DWORD dwFrameDuration = self.frameDuration;

        while (self.running) {
            llExpireTime = llExpireTime + dwFrameDuration;
            
            LONGLONG llCurrentTime = 0LL;
            xr = MWGetDeviceTime(self.hChannel, &llCurrentTime);
            if (xr != MW_SUCCEEDED) {
                llExpireTime = 0LL; usleep(10000);
                continue;
            }
            
            if (llExpireTime < llCurrentTime) {
                llExpireTime = llCurrentTime;
            }
            xr = MWScheduleTimer(self.hChannel, hTimerNotify, llExpireTime);
            if (xr != MW_SUCCEEDED) {
                llExpireTime = llCurrentTime;
                continue;
            }

            DWORD dwRet = MWWaitEvent(hTimerEvent, 1000);
            if (dwRet <= 0) {
                continue;
            }

            MWCAP_VIDEO_BUFFER_INFO videoBufferInfo;
            if (MW_SUCCEEDED != MWGetVideoBufferInfo(self.hChannel, &videoBufferInfo)) {
                continue;
            }

            MWCAP_VIDEO_SIGNAL_STATUS videoSignalStatus;
            xr = MWGetVideoSignalStatus(self.hChannel, &videoSignalStatus);
            if (xr != MW_SUCCEEDED) {
                continue;
            }

            MWCAP_VIDEO_FRAME_INFO videoFrameInfo;
            xr = MWGetVideoFrameInfo(self.hChannel, videoBufferInfo.iNewestBufferedFullFrame, &videoFrameInfo);
            if (xr != MW_SUCCEEDED)
                continue;

            RECT rcSrc = {0, 0, videoSignalStatus.cx, videoSignalStatus.cy};
            
            std::shared_ptr<PixelBufferFrame> frame(new PixelBufferFrame());
            frame->discard = (void (*)(PixelBufferFrame*, void*))(releasePixelBufferFrame);
            frame->param = (__bridge void*)self;
            
            [self.capturePixelFrameLock lock];
            if (!((std::queue<CVPixelBufferRef> *)self.capturePixelFrameQueue)->empty()) {
                frame->pixelBuffer = ((std::queue<CVPixelBufferRef> *)self.capturePixelFrameQueue)->front();
                ((std::queue<CVPixelBufferRef> *)self.capturePixelFrameQueue)->pop();
            }
            [self.capturePixelFrameLock unlock];
            
            if (frame->pixelBuffer) {
                CVPixelBufferLockBaseAddress(frame->pixelBuffer, 0);
                DWORD cbStride = (DWORD)CVPixelBufferGetBytesPerRowOfPlane(frame->pixelBuffer, 0);
                DWORD dwFrameSize = FOURCC_CalcImageSize(self.fourcc, self.width, self.height, cbStride);  // (DWORD)CVPixelBufferGetDataSize(pixelBuffer[index]);
                BYTE* byBuffer = (BYTE *)CVPixelBufferGetBaseAddressOfPlane(frame->pixelBuffer, 0);
                xr = MWCaptureVideoFrameToVirtualAddressEx(self.hChannel, MWCAP_VIDEO_FRAME_ID_NEWEST_BUFFERED, byBuffer, dwFrameSize, cbStride, FALSE, (MWCAP_PTR64)pixelBuffer,
                                                           self.fourcc, self.width, self.height,
                                                           0, 0, NULL, NULL, 0, 100, 0, 100, 0, MWCAP_VIDEO_DEINTERLACE_BLEND, MWCAP_VIDEO_ASPECT_RATIO_IGNORE, &rcSrc, NULL, 0, 0,
                                                           MWCAP_VIDEO_COLOR_FORMAT_UNKNOWN, MWCAP_VIDEO_QUANTIZATION_UNKNOWN, MWCAP_VIDEO_SATURATION_UNKNOWN);
                MWWaitEvent(hCaptureEvent, -1);
                CVPixelBufferUnlockBaseAddress(frame->pixelBuffer, 0);
                
                frame->timestamp = getTimestamp();

                [self onVideoCaptureFrame:byBuffer bufferLen:dwFrameSize pixelBufferFrame:frame];
                
                MWCAP_VIDEO_CAPTURE_STATUS captureStatus;
                xr = MWGetVideoCaptureStatus(self.hChannel, &captureStatus);
            }
        }
        
        xr = MWUnregisterTimer(self.hChannel, hTimerNotify);
        xr = MWStopVideoCapture(self.hChannel);
    } while (FALSE);

    MWCloseEvent(hCaptureEvent);
    MWCloseEvent(hTimerEvent);

    std::queue<CVPixelBufferRef> empty;
    [self.capturePixelFrameLock lock];
    std::swap(*((std::queue<CVPixelBufferRef> *)self.capturePixelFrameQueue), empty);
    [self.capturePixelFrameLock unlock];

    for (int i = 0; i < sizeof(pixelBuffer)/sizeof(pixelBuffer[0]); i++) {
        CFRelease(pixelBuffer[i]);
        pixelBuffer[i] = NULL;
    }
}

- (void)onAudioCaptureProc {
    MWCAP_PTR hAudioEvent   = MWCreateEvent();
    HNOTIFY hAudioNotify    = 0;

    MW_RESULT xr = MW_SUCCEEDED;
    do {
        xr = MWStartAudioCapture(self.hChannel);
        if (xr != MW_SUCCEEDED)
            break;

        hAudioNotify = MWRegisterNotify(self.hChannel, hAudioEvent, MWCAP_NOTIFY_AUDIO_FRAME_BUFFERED);
        if (hAudioNotify == 0)
            break;
        
        while (self.running) {
            DWORD dwRet = MWWaitEvent(hAudioEvent, 1000);
            if (dwRet <= 0)
                continue;

            ULONGLONG ullStatusBits = 0LL;
            MWGetNotifyStatus(self.hChannel, hAudioNotify, &ullStatusBits);
            if (ullStatusBits & MWCAP_NOTIFY_AUDIO_FRAME_BUFFERED) {
                do {
                    MWCAP_AUDIO_CAPTURE_FRAME audioFrame;
                    xr = MWCaptureAudioFrame(self.hChannel, &audioFrame);
                    if (xr == MW_SUCCEEDED) {
                        short asAudioSamples[MWCAP_AUDIO_SAMPLES_PER_FRAME * 2];
                        for (int i = 0; i < MWCAP_AUDIO_SAMPLES_PER_FRAME; i++) {
                            short sLeft = (short)(audioFrame.adwSamples[i * 8] >> 16);
                            short sRight = (short)(audioFrame.adwSamples[i * 8 + 4] >> 16);

                            asAudioSamples[i * 2] = sLeft;
                            asAudioSamples[i * 2 + 1] = sRight;
                        }

                        MWAudioRenderPut(self.audioRender, (uint8_t*)asAudioSamples, MWCAP_AUDIO_SAMPLES_PER_FRAME * 2 * sizeof(short));
                        
                        [self.aacEncLock lock];
                        if (self.aacEnc) {
                            self.aacEnc->encode((uint8_t*)asAudioSamples, MWCAP_AUDIO_SAMPLES_PER_FRAME * 2 * sizeof(short), getTimestamp());
                        }
                        [self.aacEncLock unlock];
                    }
                } while (xr == MW_SUCCEEDED && self.running);
            }
        }
        
        xr = MWUnregisterNotify(self.hChannel, hAudioNotify);
        xr = MWStopAudioCapture(self.hChannel);
    } while (0);

    MWCloseEvent(hAudioEvent);
}

- (void)onVideoEncodeProc {
    std::queue<std::shared_ptr<PixelBufferFrame> > *encQueue = (std::queue<std::shared_ptr<PixelBufferFrame> > *)self.encPixelFrameQueue;
    
    while (self.encoding) {
        std::shared_ptr<PixelBufferFrame> frame;
        
        [self.vtEncLock lock];
        if (!encQueue->empty()) {
            frame = encQueue->front();
        }
        [self.vtEncLock unlock];
        
        if (frame != NULL && frame->pixelBuffer) {
            if (self.vtEnc) {
                //printf("put video frame:%lld\n", frame->timestamp);
                mw_venc_put_imagebuffer(self.vtEnc, frame->pixelBuffer, frame->timestamp);
            }
            
            [self.vtEncLock lock];
            encQueue->pop();
            [self.vtEncLock unlock];
            
            //encode fps
            {
                uint64_t now = getTimestamp();
                if (self.lastEncodeTimestamp == 0) {
                    self.lastEncodeTimestamp = now;
                    self.encodeFrameCount = 0;
                } else {
                    self.encodeFrameCount++;
                }

                if (now - self.lastEncodeTimestamp >= 1000) {
                    self.encodeFps = self.encodeFrameCount * 1000.0 / (now - self.lastEncodeTimestamp);
                    self.lastEncodeTimestamp = now;
                    self.encodeFrameCount = 0;
                }
            }
        } else {
            usleep(5000);
        }
    }
    
    std::queue<std::shared_ptr<PixelBufferFrame> > empty;
    [self.vtEncLock lock];
    std::swap(*encQueue, empty);
    [self.vtEncLock unlock];
}

- (void)onVTBoxEncodeFrame:(const uint8_t**)nalus nalu_lens:(uint32_t*) nalu_lens nalu_num:(uint32_t) nalu_num frame_info:(mw_venc_frame_info_t*) frame_info {
    //printf("got encoded frames:%d, size[0]:%ld, frame_type:%d, timestamp:%llu\n", nbufs, bufs[0].len, frame_info->frame_type, frame_info->pts);
    if (!self.gotVideoTrack) {
        mw_mp4_video_info_t video_info;
        memset(&video_info, 0 , sizeof(video_info));
        video_info.codec_type = self.h265Enable ? MW_MP4_VIDEO_TYPE_HEVC: MW_MP4_VIDEO_TYPE_H264;
        video_info.timescale = 1000;
        video_info.width = self.width;
        video_info.height = self.height;
        mw_mp4_set_video(self.mp4Record, &video_info);
        
        self.gotVideoTrack = true;
    }
    
    static char startCode[4] = { 0x00, 0x00, 0x00, 0x01};
    uint32_t size = 0;
    for (int i = 0; i < nalu_num; i++) {
        //should check size
        memcpy(self.videoBufferFrame + size, startCode, sizeof(startCode)); size += sizeof(startCode);
        memcpy(self.videoBufferFrame + size, nalus[i], nalu_lens[i]);    size += nalu_lens[i];
    }
    if(mw_mp4_write_video(self.mp4Record, self.videoBufferFrame, size, (frame_info->pts - self.startRecordTime)) != MW_MP4_STATUS_SUCCESS){
        printf("write video fail\n");
    }
}

- (void)onAacEncodeFrame:(const uint8_t*)frame size:(size_t) size timestamp:(uint64_t) timestamp {
    //printf("got audio encode frames, size:%ld, timestamp:%llu\n", size, timestamp);
    if (!self.gotAudioTrack) {
        mw_mp4_audio_info_t audio_info = {
            .timescale = self.samplerate,
            .codec_type = MW_MP4_AUDIO_TYPE_ADTS_AAC,
            .sample_rate = 0,
            .channels = 0,
            .profile = 0
        };
        mw_mp4_set_audio(self.mp4Record, &audio_info);

        self.gotAudioTrack = true;
    }
    if(mw_mp4_write_audio(self.mp4Record, frame, (uint32_t)size, ((timestamp-self.startRecordTime) * self.samplerate / 1000)) != MW_MP4_STATUS_SUCCESS){
        printf("write audio fail\n");
    }
}

- (void)startCapture:(NSMenuItem *)deviceItem {
    //stop recording first
    {
        NSAlert *alert = nil;
        [self.vtEncLock lock];
        if (self.vtEnc) {
            alert = [[NSAlert alloc] init];
            [alert setMessageText:@"AVCapture"];
            [alert setInformativeText:@"Please stop recording first"];
            [alert setAlertStyle:NSWarningAlertStyle];
        }
        [self.vtEncLock unlock];
        
        if (alert) {
            [alert runModal];
            return;
        }
    }
    
    [self stopCapture];
    
    int index = (int)[deviceItem tag];
    char wPath[256] = {0};
    MWGetDevicePath(index, wPath);
    self.hChannel = MWOpenChannelByPath(wPath);
    if (self.hChannel == NULL) {
        printf("ERROR: Open channel failed !\n");
        return;
    }
    
    self.running = true;
    
    MWAudioRenderCreate(self.samplerate, self.audioChannels, self.bitsPerSample, &_audioRender);
    MWAudioRenderStart(self.audioRender);
    
    if (self.videoCaptureThreadId == 0) {
        pthread_t tid = 0;
        if (0 == pthread_create(&tid, NULL, onVideoCaptureThreadProc,  (__bridge void*)(self))) {
            self.videoCaptureThreadId = tid;
        }
    }
    
    if (self.audioCaptureThreadId == 0) {
        pthread_t tid = 0;
        if (0 == pthread_create(&tid, NULL, onAudioCaptureThreadProc,  (__bridge void*)(self))) {
           self.audioCaptureThreadId = tid;
        }
    }
    
    [deviceItem setState:NSControlStateValueOn];
    self.selectedDeviceItem = deviceItem;
    
    self.startRecordItem.enabled = true;
    self.stopRecordItem.enabled = false;
}

- (void)stopCapture {
    self.running = false;
    
    if (self.audioCaptureThreadId) {
        pthread_join(self.audioCaptureThreadId, NULL);
        self.audioCaptureThreadId = 0;
    }
    
    if (self.audioRender) {
        MWAudioRenderStop(self.audioRender);
        MWAudioRenderDestroy(self.audioRender);
        self.audioRender = NULL;
    }
    
    if (self.videoCaptureThreadId) {
        pthread_join(self.videoCaptureThreadId, NULL);
        self.videoCaptureThreadId = 0;
    }
    
    if (self.hChannel) {
        MWCloseChannel(self.hChannel);
        self.hChannel = NULL;
    }
    
    if (self.selectedDeviceItem) {
        [self.selectedDeviceItem setState:NSControlStateValueOff];
        self.selectedDeviceItem = nil;
    }
}

- (void)startRecord:(NSMenuItem *)menuItem {
    struct timeval tv = { 0 }; char date[32] = {0};
    gettimeofday(&tv, NULL); strftime(date, 32, "%Y%m%d%H%M%S", localtime(&tv.tv_sec));
    NSSavePanel*    panel = [NSSavePanel savePanel];
    [panel setNameFieldStringValue:[NSString stringWithFormat:@"%s.mp4", date]];
    [panel setMessage:@"Choose path to save record file"];
    [panel setExtensionHidden:YES];
    [panel setCanCreateDirectories:YES];
    [panel beginSheetModalForWindow:self.view.window completionHandler:^(NSInteger result){
        if (result == NSFileHandlingPanelOKButton) {
            NSString *fileName = [[panel URL] path];
            
            if (NULL == self.mp4Record) {
                   self->_mp4Record = mw_mp4_open([fileName UTF8String]);
                   
                   if (NULL == self.videoBufferFrame) {
                       self.videoBufferFrameSize = self.width * self.height * 3/2 + 512;
                       self.videoBufferFrame = (uint8_t*) malloc(self.videoBufferFrameSize);
                   }
                   
                   self.startRecordTime = getTimestamp();
               }
               
               {
                   self.encoding = true;
                   if (self.videoEncodeThreadId == 0) {
                       pthread_t tid = 0;
                       if (0 == pthread_create(&tid, NULL, onVideoEncodeThreadProc,  (__bridge void*)(self))) {
                           self.videoEncodeThreadId = tid;
                       }
                   }
                   
                   [self.vtEncLock lock];
                   
                   if (NULL == self.vtEnc) {
                       printf("start video encode\n");
                       
                       mw_venc_param_t param = {
                           .code_type = self.h265Enable ? MW_VENC_CODE_TYPE_H265 : MW_VENC_CODE_TYPE_H264,
                           .fourcc = MW_VENC_FOURCC_NV12,
                           .targetusage = MW_VENC_TARGETUSAGE_BEST_SPEED,
                           .rate_control = {
                               .mode = MW_VENC_RATECONTROL_VBR,
                               .target_bitrate = self.bitRate,
                               .max_bitrate = 0
                           },
                           .width = (int32_t)self.width,
                           .height = (int32_t)self.height,
                           .fps = {
                               .num = (int32_t)(round)(10000000/(float)self.frameDuration),
                               .den = 1
                           },
                           .slice_num = 1,
                           .gop_pic_size = (int32_t)(round)(10000000/(float)self.frameDuration),
                           .gop_ref_size = 1,
                           .profile = self.h265Enable ? MW_VENC_PROFILE_H265_MAIN : MW_VENC_PROFILE_H264_BASELINE,
                           .level = MW_VENC_LEVEL_UNKNOWN,
                           .intel_async_depth = 0,
                       };
                       
                       self.vtEnc = mw_venc_create(&param, NULL, onVTBoxEncodeCallback, (__bridge void*)self);
                       if (NULL == self.vtEnc) {
                           NSAlert *alert = [[NSAlert alloc] init];
                           [alert setMessageText:@"AVCapture"];
                           [alert setInformativeText:@"Failed to create video encoder!"];
                           [alert setAlertStyle:NSWarningAlertStyle];
                           
                           if (alert) {
                               [alert runModal];
                           }
                       }
                       
                       self.startRecordItem.enabled = !self.startRecordItem.enabled;
                       self.stopRecordItem.enabled = !self.startRecordItem.enabled;
                   }
                   
                   [self.vtEncLock unlock];
               }
               
               {
                   [self.aacEncLock lock];
                   
                   if (NULL == self.aacEnc) {
                       printf("start audio encode\n");
                       
                       self.aacEnc = new FAACEncoder();
                       self.aacEnc->open(self.samplerate, self.audioChannels);
                       self.aacEnc->setCallback(onAacEncodeCallback, (__bridge void*)self);
                   }
                   
                   [self.aacEncLock unlock];
               }
            
        }
    }];
}

- (void)stopRecord {
    self.encoding = false;
    if (self.videoEncodeThreadId) {
        pthread_join(self.videoEncodeThreadId, NULL);
        self.videoEncodeThreadId = 0;
    }
    
    [self.vtEncLock lock];
    if (self.vtEnc) {
        printf("stopping video encode\n");
        
        mw_venc_destory(self.vtEnc);
        self.vtEnc = NULL;
        
        printf("stopped video encode\n");
    }
    [self.vtEncLock unlock];
    
    [self.aacEncLock lock];
    if (self.aacEnc) {
        printf("stop audio encode\n");
        
        self.aacEnc->close();
        delete self.aacEnc;
        self.aacEnc = NULL;
    }
    [self.aacEncLock unlock];
    
    self.encodeFps           = 0.0;
    self.lastEncodeTimestamp = 0;
    self.encodeFrameCount    = 0;
    
    if (self.mp4Record) {
        mw_mp4_close(self.mp4Record);
        self.mp4Record = NULL;
        
        self.startRecordItem.enabled = !self.startRecordItem.enabled;
        self.stopRecordItem.enabled = !self.startRecordItem.enabled;
    }
    self.gotVideoTrack = false;
    self.gotAudioTrack = false;
    
    if (self.videoBufferFrame) {
        free(self.videoBufferFrame);
        self.videoBufferFrame = NULL;
    }
    self.videoBufferFrameSize = 0;
    self.startRecordTime = 0;
}

- (void)initParameters {
    self.width         = 1920;
    self.height        = 1080;
    self.frameDuration = 166667;
    self.fourcc        = MWFOURCC_NV12;
    self.pixelFormat = kCVPixelFormatType_420YpCbCr8BiPlanarVideoRange;

    self.samplerate    = 48000;
    self.audioChannels = 2;
    self.bitsPerSample = 16;

    self.audioRender   = NULL;

    self.hChannel             = NULL;
    self.capturePixelFrameLock = [[NSLock alloc] init];
    self.capturePixelFrameQueue = new std::queue<CVPixelBufferRef>();
    self.running              = false;
    self.videoCaptureThreadId = 0;
    self.audioCaptureThreadId = 0;
    self.captureFps      = 0.0;
    self.lastTimestamp   = 0;
    self.videoFrameCount = 0;
    
    self.bitRate = 4 * 1000;
    self.h265Enable = false;
    self.vtEnc = NULL;
    self.vtEncLock = [[NSLock alloc] init];
    self.encPixelFrameQueue = new std::queue<std::shared_ptr<PixelBufferFrame> >();
    self.encoding = false;
    self.videoEncodeThreadId = 0;
    self.encodeFps           = 0.0;
    self.lastEncodeTimestamp = 0;
    self.encodeFrameCount    = 0;
    
    self.aacEnc = NULL;
    self.aacEncLock = [[NSLock alloc] init];
    
    self.mp4Record = NULL;
    self.gotVideoTrack = false;
    self.gotAudioTrack = false;
    self.videoBufferFrame = NULL;
    self.videoBufferFrameSize = 0;
    self.startRecordTime = 0;
}

- (void)initDeviceMenu {
    int count = MWGetChannelCount();
    for (int i = 0; i < count; i++) {
        NSString *itemTitle = nil;
        
        MWCAP_CHANNEL_INFO mci = { 0 };
        MWGetChannelInfoByIndex(i, &mci);
        if (strcmp(mci.szFamilyName, "Pro Capture") == 0) {
            itemTitle = [NSString stringWithFormat:@"%02d-%d %s", mci.byBoardIndex, mci.byChannelIndex, mci.szProductName];
            NSMenuItem *deviceItem = [self.deviceMenu addItemWithTitle:itemTitle action:@selector(startCapture:) keyEquivalent:@""];
            [deviceItem setTag:i];
        }
    }
}

- (void)createPreviewLayer {
    if (!self.videoLayer) {
        self.videoLayer = [[AVSampleBufferDisplayLayer alloc] init];
        self.videoLayer.frame = CGRectMake(0, self.statusLabel.frame.size.height,
                                           self.view.frame.size.width, self.view.frame.size.height - self.statusLabel.frame.size.height);
        self.videoLayer.bounds = CGRectMake(0, self.statusLabel.frame.size.height,
                                            self.view.bounds.size.width, self.view.bounds.size.height - self.statusLabel.frame.size.height);
        self.videoLayer.autoresizingMask = NSViewWidthSizable | NSViewHeightSizable;
        self.videoLayer.videoGravity = AVLayerVideoGravityResizeAspect;
        self.videoLayer.backgroundColor = [NSColor blackColor].CGColor;
        [[self.view layer] addSublayer:self.videoLayer];
        
        self.view.window.delegate = self;
    }
}

- (void)parseArgs {
    NSArray *args = [[NSProcessInfo processInfo] arguments];
    for (int i=0; i<args.count; i++) {
        NSString *item = [args objectAtIndex:i];
        const char *ptr = strstr([item UTF8String], "-w=");
        if (ptr) {
            self.width = atoi(ptr+strlen("-w="));
            continue;
        }
        
        ptr = strstr([item UTF8String], "-h=");
        if (ptr) {
            self.height = atoi(ptr+strlen("-h="));
            continue;
        }
        
        ptr = strstr([item UTF8String], "-f=");
        if (ptr) {
            self.frameDuration = atoi(ptr+strlen("-f="));
            continue;
        }
        
        ptr = strstr([item UTF8String], "-v=");
        if (ptr) {
           self.viewEnable = atoi(ptr+strlen("-v="));
           continue;
        }
        
        ptr = strstr([item UTF8String], "-c=");
        if (ptr) {
           self.h265Enable = (0 == strcmp("h265", ptr+strlen("-c=")));
           continue;
        }
        
        ptr = strstr([item UTF8String], "-b=");
        if (ptr) {
            self.bitRate = atoi(ptr+strlen("-b="));
            continue;
        }
    }
}

- (void)viewDidLoad {
    [self.view setWantsLayer:YES];
    [super viewDidLoad];
    // Do any additional setup after loading the view.
    
    //base UI
    {
        self.viewEnable = 1;
        [self performSelector:@selector(createPreviewLayer) withObject:nil afterDelay:0];
        
        NSMenu *mainMenu = [NSApp mainMenu];
        
        NSMenuItem *mainMenuFileItem = [[NSMenuItem alloc] init];
        [mainMenu addItem:mainMenuFileItem];
        NSMenu *fileMenu = [[NSMenu alloc] initWithTitle:@"File"];
        [mainMenuFileItem setSubmenu:fileMenu];
        [fileMenu setAutoenablesItems:NO];
        self.startRecordItem = [fileMenu addItemWithTitle:@"Start record" action:@selector(startRecord:) keyEquivalent:@""];
        self.startRecordItem.enabled = false;
        self.stopRecordItem = [fileMenu addItemWithTitle:@"Stop record" action:@selector(stopRecord) keyEquivalent:@""];
        self.stopRecordItem.enabled = false;
        
        NSMenuItem *mainMenuDeviceItem = [[NSMenuItem alloc] init];
        [mainMenu addItem:mainMenuDeviceItem];
        self.deviceMenu = [[NSMenu alloc] initWithTitle:@"Device"];
        [mainMenuDeviceItem setSubmenu:self.deviceMenu];
    }
    
    MWCaptureInitInstance();
    MWRefreshDevice();
    
    [self initParameters]; [self parseArgs];
    [self initDeviceMenu];
    
    if (self.deviceMenu.numberOfItems > 0) {
        [self performSelector:@selector(startCapture:) withObject:[self.deviceMenu itemAtIndex:0] afterDelay:0];
    }
    
    self.statusTimer = [NSTimer scheduledTimerWithTimeInterval:0.5 target: self selector:@selector(statusTimerProc:) userInfo:nil repeats:YES];
}

- (BOOL)windowShouldClose:(NSWindow *)sender {
    [self stopRecord];
    [self stopCapture];
    
    if (self.statusTimer) {
        [self.statusTimer invalidate];
    }
    
    std::queue<std::shared_ptr<PixelBufferFrame> > *endQueue = (std::queue<std::shared_ptr<PixelBufferFrame> > *)self.encPixelFrameQueue;
    delete endQueue;
    self.encPixelFrameQueue = NULL;
    
    std::queue<CVPixelBufferRef> *captureQueue = (std::queue<CVPixelBufferRef> *)self.capturePixelFrameQueue;
    delete captureQueue;
    self.capturePixelFrameQueue = NULL;
    
    MWCaptureExitInstance();
    
    return YES;
}

- (void)viewDidDisappear {
}

- (void)setRepresentedObject:(id)representedObject {
    [super setRepresentedObject:representedObject];

    // Update the view, if already loaded.
}

- (void) statusTimerProc: (NSTimer *) timer {
    char strFourcc[5] = { 0 };
    *(uint32_t*)strFourcc = self.fourcc;
    [self.statusLabel setStringValue:[NSString stringWithFormat:@" %d x %d,    %s,    Cap:%.2f,    Enc: %.2f ", self.width, self.height, strFourcc, self.captureFps, self.encodeFps]];
}

@end
